# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from six.moves import queue

from orquesta.composers import base as comp_base
from orquesta import graphing
from orquesta.specs import native as native_specs


LOG = logging.getLogger(__name__)


class WorkflowComposer(comp_base.WorkflowComposer):
    wf_spec_type = native_specs.WorkflowSpec

    @classmethod
    def compose(cls, spec):
        if not cls.wf_spec_type:
            raise TypeError("Undefined spec type for composer.")

        if not isinstance(spec, cls.wf_spec_type):
            raise TypeError('Unsupported spec type "%s".' % str(type(spec)))

        return cls._compose_wf_graph(spec)

    @classmethod
    def _compose_wf_graph(cls, wf_spec):
        if not isinstance(wf_spec, cls.wf_spec_type):
            raise TypeError("Workflow spec is not typeof %s." % cls.wf_spec_type.__name__)

        q = queue.Queue()
        wf_graph = graphing.WorkflowGraph()

        for task_name, condition, task_transition_item_idx in wf_spec.tasks.get_start_tasks():
            q.put((task_name, []))

        while not q.empty():
            task_name, splits = q.get()

            wf_graph.add_task(task_name)

            if wf_spec.tasks.is_join_task(task_name):
                task_spec = wf_spec.tasks[task_name]
                barrier = "*" if task_spec.join == "all" else task_spec.join
                wf_graph.set_barrier(task_name, value=barrier)

            # Determine if the task is a split task and if it is in a cycle. If the task is a
            # split task, keep track of where the split(s) occurs.
            if wf_spec.tasks.is_split_task(task_name) and not wf_spec.tasks.in_cycle(task_name):
                splits.append(task_name)

            if splits:
                wf_graph.update_task(task_name, splits=splits)

            # Update task attributes if task spec has retry criteria.
            task_spec = wf_spec.tasks.get_task(task_name)

            if task_spec.has_retry():
                retry_spec = {
                    "when": getattr(task_spec.retry, "when", None),
                    "count": getattr(task_spec.retry, "count", None),
                    "delay": getattr(task_spec.retry, "delay", None),
                }

                wf_graph.update_task(task_name, retry=retry_spec)

            # Add task transition to the workflow graph.
            next_tasks = wf_spec.tasks.get_next_tasks(task_name)

            for next_task_name, condition, task_transition_item_idx in next_tasks:
                if next_task_name == "retry":
                    retry_spec = {"when": condition or "<% completed() %>", "count": 3}
                    wf_graph.update_task(task_name, retry=retry_spec)
                    continue

                if not wf_graph.has_task(next_task_name) or not wf_spec.tasks.in_cycle(
                    next_task_name
                ):
                    q.put((next_task_name, list(splits)))

                crta = [condition] if condition else []

                seqs = wf_graph.has_transition(
                    task_name, next_task_name, criteria=crta, ref=task_transition_item_idx
                )

                # Use existing transition if present otherwise create new transition.
                if seqs:
                    wf_graph.update_transition(
                        task_name,
                        next_task_name,
                        key=seqs[0][2],
                        criteria=crta,
                        ref=task_transition_item_idx,
                    )
                else:
                    wf_graph.add_transition(
                        task_name, next_task_name, criteria=crta, ref=task_transition_item_idx
                    )

        return wf_graph



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=no-member

import datetime
import logging
import re
import six

import dateutil.parser


LOG = logging.getLogger(__name__)


ISO8601_FORMAT = r"%Y-%m-%dT%H:%M:%S"
ISO8601_FORMAT_USEC = ISO8601_FORMAT + r".%f"
ISO8601_REGEX = r"\d{4}\-\d{2}\-\d{2}(\s|T)\d{2}:\d{2}:\d{2}(\.\d{3,6})"
ISO8601_REGEX_NAIVE = r"^%s$" % ISO8601_REGEX
ISO8601_REGEX_UTC = r"^%s?(Z|\+00|\+0000|\+00:00)$" % ISO8601_REGEX
ISO8601_REGEX_TZ = r"^%s?(Z|\+\d{2}|\+\d{4}|\+\d{2}:\d{2})$" % ISO8601_REGEX


def valid(v):
    is_dt_obj = isinstance(v, datetime.datetime)
    is_dt_str = isinstance(v, six.string_types) and re.match(ISO8601_REGEX_NAIVE, v)
    is_dt_str_tz = isinstance(v, six.string_types) and re.match(ISO8601_REGEX_TZ, v)

    return is_dt_obj or is_dt_str or is_dt_str_tz


def format(dt, usec=True, offset=True):
    if isinstance(dt, six.string_types):
        dt = parse(dt)

    fmt = ISO8601_FORMAT_USEC if usec else ISO8601_FORMAT

    if offset:
        ost = dt.strftime("%z")
        ost = (ost[:3] + ":" + ost[3:]) if ost else "+00:00"
    else:
        tz = dt.tzinfo.tzname(dt) if dt.tzinfo else "UTC"
        ost = "Z" if tz == "UTC" else tz

    return dt.strftime(fmt) + ost


def parse(v):
    return dateutil.parser.parse(str(v))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.specs import loader as spec_loader
from orquesta.specs import mistral as mistral_specs
from orquesta.specs.mistral import v2 as mistral_v2_specs


class SpecTest(unittest.TestCase):
    def setUp(self):
        super(SpecTest, self).setUp()
        self.spec_module_name = "mistral"

    def test_get_module(self):
        self.assertEqual(spec_loader.get_spec_module(self.spec_module_name), mistral_specs)

    def test_get_spec(self):
        spec_module = spec_loader.get_spec_module(self.spec_module_name)

        self.assertEqual(spec_module.WorkflowSpec, mistral_specs.WorkflowSpec)

    def test_spec_catalog(self):
        spec_module = spec_loader.get_spec_module(self.spec_module_name)

        self.assertEqual(spec_module.WorkflowSpec.get_catalog(), self.spec_module_name)

    def test_spec_version(self):
        self.assertEqual("2.0", mistral_v2_specs.VERSION)
        self.assertEqual("2.0", mistral_specs.VERSION)

    def test_workflow_spec_imports(self):
        self.assertEqual(mistral_specs.WorkflowSpec, mistral_v2_specs.workflows.WorkflowSpec)

    def test_task_spec_imports(self):
        self.assertEqual(mistral_specs.TaskDefaultsSpec, mistral_v2_specs.tasks.TaskDefaultsSpec)

        self.assertEqual(mistral_specs.TaskSpec, mistral_v2_specs.tasks.TaskSpec)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.specs.mistral import base as test_base


class WorkflowSpecVarsValidationTest(test_base.MistralWorkflowSpecTest):
    def test_bad_vars_yaql(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                input:
                    - y
                vars:
                    foo: <% ctx().a %>
                    fooa: <% ctx().x + ctx().y + ctx().z %>
                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().a %>",
                    "message": 'Variable "a" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().x + ctx().y + ctx().z %>",
                    "message": 'Variable "x" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().x + ctx().y + ctx().z %>",
                    "message": 'Variable "z" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_jinja(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                input:
                    - y
                vars:
                    foo: "{{ ctx().a }}"
                    fooa: "{{ ctx().x + ctx().y + ctx().z }}"
                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "jinja",
                    "expression": "{{ ctx().a }}",
                    "message": 'Variable "a" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
                {
                    "type": "jinja",
                    "expression": "{{ ctx().x + ctx().y + ctx().z }}",
                    "message": 'Variable "x" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
                {
                    "type": "jinja",
                    "expression": "{{ ctx().x + ctx().y + ctx().z }}",
                    "message": 'Variable "z" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_mix_yaql_and_jinja(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                input:
                    - y
                vars:
                    foo: "{{ ctx().a }}"
                    fooa: <% ctx().x + ctx().y + ctx().z %>
                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "jinja",
                    "expression": "{{ ctx().a }}",
                    "message": 'Variable "a" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().x + ctx().y + ctx().z %>",
                    "message": 'Variable "x" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().x + ctx().y + ctx().z %>",
                    "message": 'Variable "z" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "sequential.vars",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_in_sequential_tasks(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                vars:
                    foobar: foobar
                tasks:
                    task1:
                        action: std.echo
                        input:
                            message: <% ctx().foobar %>
                        on-success:
                            - task2
                    task2:
                        action: std.echo
                        input:
                            message: <% ctx().foo %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.properties.task2.properties.input",
                    "spec_path": "sequential.tasks.task2.input",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_in_branching_tasks(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                vars:
                    foobar: foobar
                tasks:
                    task1:
                        action: std.echo
                        input:
                            message: <% ctx().foobar %>
                        publish:
                            foo: bar
                        on-success:
                            - task2
                    task2:
                        action: std.echo
                        input:
                            message: <% ctx().foo + ctx().bar %>

                    task3:
                        action: std.echo
                        input:
                            message: <% ctx().foobar %>
                        publish:
                            bar: foo
                        on-success:
                            - task4
                    task4:
                        action: std.echo
                        input:
                            message: <% ctx().foo + ctx().bar %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "bar" is referenced before assignment.',
                    "schema_path": "properties.tasks.properties.task2.properties.input",
                    "spec_path": "sequential.tasks.task2.input",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.properties.task4.properties.input",
                    "spec_path": "sequential.tasks.task4.input",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_in_reusable_split_tasks(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                vars:
                    foobar: foobar
                tasks:
                    task1:
                        action: std.echo
                        input:
                            message: <% ctx().foobar %>
                        publish:
                            foo: bar
                        on-success:
                            - task2
                    task2:
                        action: std.echo
                        input:
                            message: <% ctx().foo + ctx().bar %>
                        on-success:
                            - task5

                    task3:
                        action: std.echo
                        input:
                            message: <% ctx().foobar %>
                        publish:
                            bar: foo
                        on-success:
                            - task4
                    task4:
                        action: std.echo
                        input:
                            message: <% ctx().foo + ctx().bar %>
                        on-success:
                            - task5

                    task5:
                        action: std.echo
                        input:
                            message: <% ctx().foo %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "bar" is referenced before assignment.',
                    "schema_path": "properties.tasks.properties.task2.properties.input",
                    "spec_path": "sequential.tasks.task2.input",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.properties.task4.properties.input",
                    "spec_path": "sequential.tasks.task4.input",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.properties.task5.properties.input",
                    "spec_path": "sequential.tasks.task5.input",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_vars_in_join_task(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A complex branching workflow.
                vars:
                    foobar: foobar
                tasks:
                    task1:
                        action: std.echo
                        input:
                            message: <% ctx().foobar %>
                        publish:
                            foo: bar
                        on-success:
                            - task2
                    task2:
                        action: std.echo
                        input:
                            message: <% ctx().foo %>
                        on-success:
                            - task5

                    task3:
                        action: std.echo
                        input:
                            message: <% ctx().foobar %>
                        publish:
                            bar: foo
                        on-success:
                            - task4
                    task4:
                        action: std.echo
                        input:
                            message: <% ctx().bar %>
                        on-success:
                            - task5

                    task5:
                        join: all
                        action: std.echo
                        input:
                            message: <% ctx().foo + ctx().bar %>
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), dict())



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorPausePendingResumeTest(test_base.WorkflowConductorTest):
    def test_pause_and_resume_from_workflow(self):
        wf_def = """
        version: 1.0
        description: A basic branching workflow.
        tasks:
          # branch 1
          task1:
            action: core.ask
            next:
              - when: <% succeeded() %>
                do: task3
          # branch 2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3
          # adjoining branch
          task3:
            join: all
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Run task1 and task2.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSING)

        # Put task1 to pending state and assert workflow is pausing.
        self.forward_task_statuses(conductor, "task1", [statuses.PENDING])
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        # Complete task1 only. The workflow should still be pausing
        # because task2 is still running.
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        # Complete task2. When task2 completes, the workflow should be paused
        # because there is no task in active status.
        self.forward_task_statuses(conductor, "task2", [statuses.SUCCEEDED])
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the workflow, task3 should be staged, and complete task3.
        conductor.request_workflow_status(statuses.RESUMING)
        self.assert_next_task(conductor, "task3", {})
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING, statuses.SUCCEEDED])
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_pause_and_resume_from_branches(self):
        wf_def = """
        version: 1.0
        description: A basic branching workflow.
        tasks:
          # branch 1
          task1:
            action: core.ask
            next:
              - when: <% succeeded() %>
                do: task3
          # branch 2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3
          # adjoining branch
          task3:
            join: all
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Run task1 and task2.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Pause task1 and task2.
        self.forward_task_statuses(conductor, "task1", [statuses.PENDING])
        self.forward_task_statuses(conductor, "task2", [statuses.PAUSED])
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume and complete task1 only. Once task1 completes, the workflow
        # should pause again because there is no active task.
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume and complete task2. When task2 completes, the workflow
        # should stay running because task3 is now staged and ready.
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.forward_task_statuses(conductor, "task2", [statuses.SUCCEEDED])
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.assert_next_task(conductor, "task3", {})

        # Complete task3.
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING, statuses.SUCCEEDED])
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_pending_and_failed_with_task_transition_error(self):
        wf_def = """
        version: 1.0
        description: A basic sequential workflow.
        tasks:
          task1:
            action: core.ask
            next:
              - when: <% result().response.foobar %>
                do: task2
          task2:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Run task1.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Put task1 to pending state and assert workflow is paused.
        self.forward_task_statuses(conductor, "task1", [statuses.PENDING])
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Complete task1 and assert workflow execution fails
        # due to the expression error in task transition.
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import logging
import six

from orquesta.utils import plugin as plugin_util


LOG = logging.getLogger(__name__)


def get_composer(catalog):
    return plugin_util.get_module("orquesta.composers", catalog)


@six.add_metaclass(abc.ABCMeta)
class WorkflowComposer(object):
    wf_spec_type = None

    @classmethod
    @abc.abstractmethod
    def compose(cls, spec):
        raise NotImplementedError()



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imp
import inspect
import sys
import traceback

from hacking import core


MODULES_CACHE = dict((mod, True) for mod in tuple(sys.modules.keys()) + sys.builtin_module_names)


@core.flake8ext
def check_module_only(logical_line, physical_line, filename, noqa):
    """Check imports.

    Imports should be modules only.

    Examples:
    Okay: from orquesta import conducting
    Okay: from orquesta import states
    Okay: from orquesta.utils import date
    O101: from orquesta.conducting import WorkflowConductor
    O101: from orquesta.states import REQUESTED
    O101: from orquesta.utils.date import parse
    """

    if noqa:
        return

    def check_is_module(mod, search_path=sys.path):
        mod = mod.replace("(", "")  # Ignore parentheses

        for finder in sys.meta_path:
            if finder.find_module(mod) is not None:
                return True

        try:
            mod_name = mod

            while "." in mod_name:
                pack_name, _sep, mod_name = mod_name.partition(".")
                f, p, d = imp.find_module(pack_name, search_path)
                search_path = [p]

            imp.find_module(mod_name, search_path)
        except ImportError:
            try:
                # Handle namespace modules
                if "." in mod:
                    pack_name, mod_name = mod.rsplit(".", 1)
                    __import__(pack_name, fromlist=[mod_name])
                else:
                    __import__(mod)
            except ImportError:
                # Import error here means the thing is not importable in
                # current environment, either because of missing dependency,
                # typo in code being checked, or any other reason. Anyway, we
                # have no means to know if it is module or not, so we return
                # True to avoid false positives.
                return True
            except Exception:
                # Skip stack trace on unexpected import error,
                # log and continue.
                traceback.print_exc()
                return False
            else:
                # The line was imported. If it is a module, it must be there.
                # If it is not there under its own name, look to see if it is
                # a module redirection import..
                if mod in sys.modules:
                    return True
                else:
                    pack_name, _sep, mod_name = mod.rpartition(".")
                    if pack_name in sys.modules:
                        _mod = getattr(sys.modules[pack_name], mod_name, None)
                        return inspect.ismodule(_mod)
                    return False

        return True

    def is_module(mod):
        if mod not in MODULES_CACHE:
            MODULES_CACHE[mod] = check_is_module(mod)

        return MODULES_CACHE[mod]

    parts = logical_line.split()

    if (
        len(parts) > 3
        and parts[0] == "from"
        and parts[1] != "__future__"
        and not core.is_import_exception(parts[1])
        and not is_module(".".join([parts[1], parts[3]]))
    ):
        yield 0, "O101 import must be module only"



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorWithItemsPauseResumeTest(test_base.WorkflowConductorWithItemsTest):
    def test_pause_item_list_processed(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.SUCCEEDED,
            statuses.PAUSED,
            statuses.SUCCEEDED,
            statuses.SUCCEEDED,
        ]

        expected_task_statuses = [
            statuses.RUNNING,
            statuses.PAUSING,
            statuses.PAUSING,
            statuses.PAUSED,
        ]

        expected_workflow_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
        ]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow is paused.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

    def test_pause_item_list_incomplete(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED, statuses.PAUSED, statuses.SUCCEEDED]
        expected_task_statuses = [statuses.RUNNING, statuses.PAUSING, statuses.PAUSED]
        expected_workflow_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.PAUSED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow is canceled.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

    def test_resume_paused_item_list_processed(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.SUCCEEDED,
            statuses.PAUSED,
            statuses.SUCCEEDED,
            statuses.SUCCEEDED,
        ]

        expected_task_statuses = [
            statuses.RUNNING,
            statuses.PAUSING,
            statuses.PAUSING,
            statuses.PAUSED,
        ]

        expected_workflow_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
        ]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the paued action execution.
        self.forward_task_item_statuses(conductor, task_name, 1, [statuses.RUNNING])

        # Assert the task and workflow is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)

        staged_task = conductor.workflow_state.get_staged_task(task_name, task_route)
        self.assertIsNotNone(staged_task)
        self.assertIn("items", staged_task)
        self.assertEqual(staged_task["items"][1]["status"], statuses.RUNNING)

        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Complete the resumed action execution.
        self.forward_task_item_statuses(
            conductor, task_name, 1, [statuses.SUCCEEDED], result=task_ctx["xs"][1]
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the task and workflow succeeded.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.SUCCEEDED)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_pause_workflow_using_pausing_status_with_items_active(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        # Verify the set of action executions.
        expected_task = self.format_task_item(
            task_name,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_name),
            actions=task_action_specs,
            items_count=len(task_ctx["xs"]),
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # Set the items to running status.
        for i in range(0, len(task_ctx["xs"])):
            self.forward_task_item_statuses(conductor, task_name, i, [statuses.RUNNING])

        # Assert that the task is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.PAUSING)

        # Complete the items.
        for i in range(0, len(task_ctx["xs"])):
            result = task_ctx["xs"][i]
            status_changes = [statuses.SUCCEEDED]
            self.forward_task_item_statuses(conductor, task_name, i, status_changes, result=result)

        # Assert the task and workflow are completed.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.SUCCEEDED)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the workflow.
        conductor.request_workflow_status(statuses.RESUMING)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_pause_workflow_using_paused_status_with_items_active(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        # Verify the set of action executions.
        expected_task = self.format_task_item(
            task_name,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_name),
            actions=task_action_specs,
            items_count=len(task_ctx["xs"]),
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # Set the items to running status.
        for i in range(0, len(task_ctx["xs"])):
            self.forward_task_item_statuses(conductor, task_name, i, [statuses.RUNNING])

        # Assert that the task is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSED)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.PAUSING)

        # Complete the items.
        for i in range(0, len(task_ctx["xs"])):
            result = task_ctx["xs"][i]
            status_changes = [statuses.SUCCEEDED]
            self.forward_task_item_statuses(conductor, task_name, i, status_changes, result=result)

        # Assert the task and workflow are completed.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.SUCCEEDED)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the workflow.
        conductor.request_workflow_status(statuses.RESUMING)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_pause_workflow_using_pausing_status_with_items_concurrency(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with:
              items: <% ctx(xs) %>
              concurrency: 2
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        concurrency = 2

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED] * 2
        expected_task_statuses = [statuses.RUNNING] * 2
        expected_workflow_statuses = [statuses.RUNNING] * 2

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
            concurrency=concurrency,
        )

        # Assert the task is not removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow is still running.
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.PAUSED)

        # Resume the workflow.
        conductor.request_workflow_status(statuses.RESUMING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)

        # Verify the second set of action executions.
        expected_task = self.format_task_item(
            task_name,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_name),
            actions=task_action_specs[concurrency:],
            items_count=len(task_ctx["xs"]),
            items_concurrency=concurrency,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # Set the items to running status.
        for i in range(0 + concurrency, len(task_ctx["xs"])):
            self.forward_task_item_statuses(conductor, task_name, i, [statuses.RUNNING])

        # Assert that the task is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Complete the items.
        for i in range(0 + concurrency, len(task_ctx["xs"])):
            result = task_ctx["xs"][i]
            status_changes = [statuses.SUCCEEDED]
            self.forward_task_item_statuses(conductor, task_name, i, status_changes, result=result)

        # Assert the task and workflow are completed.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.SUCCEEDED)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_pause_workflow_using_paused_status_with_items_concurrency(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with:
              items: <% ctx(xs) %>
              concurrency: 2
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        concurrency = 2

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED] * 2
        expected_task_statuses = [statuses.RUNNING] * 2
        expected_workflow_statuses = [statuses.RUNNING] * 2

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
            concurrency=concurrency,
        )

        # Assert the task is not removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow is still running.
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSED)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.PAUSED)

        # Resume the workflow.
        conductor.request_workflow_status(statuses.RESUMING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)

        # Verify the second set of action executions.
        expected_task = self.format_task_item(
            task_name,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_name),
            actions=task_action_specs[concurrency:],
            items_count=len(task_ctx["xs"]),
            items_concurrency=concurrency,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # Set the items to running status.
        for i in range(0 + concurrency, len(task_ctx["xs"])):
            self.forward_task_item_statuses(conductor, task_name, i, [statuses.RUNNING])

        # Assert that the task is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Complete the items.
        for i in range(0 + concurrency, len(task_ctx["xs"])):
            result = task_ctx["xs"][i]
            status_changes = [statuses.SUCCEEDED]
            self.forward_task_item_statuses(conductor, task_name, i, status_changes, result=result)

        # Assert the task and workflow are completed.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.SUCCEEDED)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_pause_workflow_with_items_concurrency_and_active(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with:
              items: <% ctx(xs) %>
              concurrency: 2
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        concurrency = 2

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        # Verify the first set of action executions.
        expected_task = self.format_task_item(
            task_name,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_name),
            actions=task_action_specs[0:concurrency],
            items_count=len(task_ctx["xs"]),
            items_concurrency=concurrency,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # Set the items to running status.
        for i in range(0, concurrency):
            self.forward_task_item_statuses(conductor, task_name, i, [statuses.RUNNING])

        # Assert that the task is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.PAUSING)

        # Complete the items.
        for i in range(0, concurrency):
            result = task_ctx["xs"][i]
            status_changes = [statuses.SUCCEEDED]
            self.forward_task_item_statuses(conductor, task_name, i, status_changes, result=result)

        # Assert the task and workflow are paused.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.PAUSED)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the workflow.
        conductor.request_workflow_status(statuses.RESUMING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)

        # Verify the second set of action executions.
        expected_task = self.format_task_item(
            task_name,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_name),
            actions=task_action_specs[concurrency:],
            items_count=len(task_ctx["xs"]),
            items_concurrency=concurrency,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # Set the items to running status.
        for i in range(0 + concurrency, len(task_ctx["xs"])):
            self.forward_task_item_statuses(conductor, task_name, i, [statuses.RUNNING])

        # Assert that the task is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Complete the items.
        for i in range(0 + concurrency, len(task_ctx["xs"])):
            result = task_ctx["xs"][i]
            status_changes = [statuses.SUCCEEDED]
            self.forward_task_item_statuses(conductor, task_name, i, status_changes, result=result)

        # Assert the task and workflow are completed.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.SUCCEEDED)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_pending_item_list_processed(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.SUCCEEDED,
            statuses.PENDING,
            statuses.SUCCEEDED,
            statuses.SUCCEEDED,
        ]

        expected_task_statuses = [
            statuses.RUNNING,
            statuses.PAUSING,
            statuses.PAUSING,
            statuses.PAUSED,
        ]

        expected_workflow_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
        ]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow is paused.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

    def test_resume_pending_item_list_processed(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.SUCCEEDED,
            statuses.PENDING,
            statuses.SUCCEEDED,
            statuses.SUCCEEDED,
        ]

        expected_task_statuses = [
            statuses.RUNNING,
            statuses.PAUSING,
            statuses.PAUSING,
            statuses.PAUSED,
        ]

        expected_workflow_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
        ]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the paued action execution.
        self.forward_task_item_statuses(conductor, task_name, 1, [statuses.RUNNING])

        # Assert the task and workflow is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)

        staged_task = conductor.workflow_state.get_staged_task(task_name, task_route)
        self.assertEqual(staged_task["items"][1]["status"], statuses.RUNNING)

        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Complete the resumed action execution.
        self.forward_task_item_statuses(
            conductor, task_name, 1, [statuses.SUCCEEDED], result=task_ctx["xs"][1]
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the task and workflow succeeded.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.SUCCEEDED)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_resume_partial(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.SUCCEEDED,
            statuses.PAUSED,
            statuses.PAUSED,
            statuses.SUCCEEDED,
        ]

        expected_task_statuses = [
            statuses.RUNNING,
            statuses.PAUSING,
            statuses.PAUSING,
            statuses.PAUSED,
        ]

        expected_workflow_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
        ]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the paued action execution.
        self.forward_task_item_statuses(conductor, task_name, 1, [statuses.RUNNING])

        # Assert the task and workflow is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)

        staged_task = conductor.workflow_state.get_staged_task(task_name, task_route)
        self.assertEqual(staged_task["items"][1]["status"], statuses.RUNNING)

        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Complete the resumed action execution.
        self.forward_task_item_statuses(
            conductor, task_name, 1, [statuses.SUCCEEDED], result=task_ctx["xs"][1]
        )

        # Assert the task is removed from staging.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the task and workflow is paused.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.PAUSED)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.composers import native as native_comp
from orquesta.tests.unit.composition.native import base as native_comp_test_base
from orquesta.utils import plugin as plugin_util


class WorkflowComposerTest(native_comp_test_base.OrchestraWorkflowComposerTest):
    def test_get_composer(self):
        self.assertEqual(
            plugin_util.get_module("orquesta.composers", self.spec_module_name),
            native_comp.WorkflowComposer,
        )



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.tests.fixtures import loader as fixture_loader


class FixtureTest(unittest.TestCase):
    def test_fixture_loader(self):
        expected = {"version": 1.0, "basic": {"tasks": [{"task1": "foobar"}]}}

        content = fixture_loader.get_fixture_content("mock/basic.yaml", "workflows")

        self.assertIsInstance(content, dict)
        self.assertDictEqual(content, expected)

    def test_fixture_fixture_loader_raw(self):
        expected = "version: 1.0\nbasic:\n  tasks:\n    - task1: foobar\n"
        content = fixture_loader.get_fixture_content("mock/basic.yaml", "workflows", raw=True)

        self.assertIsInstance(content, str)
        self.assertEqual(content, expected)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit import base as test_base
from orquesta.utils import jsonify as json_util


MOCK_LIST_DATA = {
    "vms": [
        {"name": "vmweb1", "region": "us-east", "role": "web"},
        {"name": "vmdb1", "region": "us-east", "role": "db"},
        {"name": "vmweb2", "region": "us-west", "role": "web"},
        {"name": "vmdb2", "region": "us-west", "role": "db"},
    ]
}

MOCK_DICT_DATA = {
    "vms": {
        "vmdb1": {"region": "us-east", "role": "db", "name": "vmdb1"},
        "vmdb2": {"region": "us-west", "role": "db", "name": "vmdb2"},
        "vmweb1": {"region": "us-east", "role": "web", "name": "vmweb1"},
        "vmweb2": {"region": "us-west", "role": "web", "name": "vmweb2"},
    }
}


class YAQLContextQueryTest(test_base.ExpressionEvaluatorTest):
    @classmethod
    def setUpClass(cls):
        cls.language = "yaql"
        super(YAQLContextQueryTest, cls).setUpClass()

    def test_ctx_list_query(self):
        expected_result = [i["name"] for i in MOCK_LIST_DATA["vms"]]

        expr = "<% ctx(vms).select($.name) %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertListEqual(self.evaluator.evaluate(expr, MOCK_LIST_DATA), expected_result)

        expr = "<% ctx().vms.select($.name) %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertListEqual(self.evaluator.evaluate(expr, MOCK_LIST_DATA), expected_result)

    def test_ctx_list_indexing(self):
        expected_result = {"name": "vmweb1", "region": "us-east", "role": "web"}

        expr = "<% ctx(vms)[0] %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertDictEqual(self.evaluator.evaluate(expr, MOCK_LIST_DATA), expected_result)

        expr = "<% ctx().vms[0] %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertDictEqual(self.evaluator.evaluate(expr, MOCK_LIST_DATA), expected_result)

    def test_ctx_list_to_dict_transform(self):
        expected_result = json_util.deepcopy(MOCK_DICT_DATA)

        expr = "<% dict(vms=>dict(ctx(vms).select([$.name, $]))) %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertDictEqual(self.evaluator.evaluate(expr, MOCK_LIST_DATA), expected_result)

        expr = "<% dict(vms=>dict(ctx().vms.select([$.name, $]))) %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertDictEqual(self.evaluator.evaluate(expr, MOCK_LIST_DATA), expected_result)

    def test_ctx_dict_get(self):
        expected_result = {"name": "vmweb1", "region": "us-east", "role": "web"}

        expr = "<% ctx(vms).get(vmweb1) %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertDictEqual(self.evaluator.evaluate(expr, MOCK_DICT_DATA), expected_result)

        expr = "<% ctx().vms.get(vmweb1) %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertDictEqual(self.evaluator.evaluate(expr, MOCK_DICT_DATA), expected_result)

    def test_ctx_dict_indexing(self):
        expected_result = {"name": "vmweb1", "region": "us-east", "role": "web"}

        expr = "<% ctx(vms)[vmweb1] %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertDictEqual(self.evaluator.evaluate(expr, MOCK_DICT_DATA), expected_result)

        expr = "<% ctx().vms[vmweb1] %>"
        self.assertListEqual([], self.evaluator.validate(expr))
        self.assertDictEqual(self.evaluator.evaluate(expr, MOCK_DICT_DATA), expected_result)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import logging

import networkx as nx
from networkx.readwrite import json_graph

import six

from orquesta import exceptions as exc
from orquesta.utils import dictionary as dict_util
from orquesta.utils import jsonify as json_util


LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class WorkflowGraph(object):
    def __init__(self, graph=None):
        # self._graph is the graph model for the workflow. The tracking of workflow and task
        # progress and state is separate from the graph model. There are use cases where tasks
        # may be cycled and states overwritten.
        self._graph = graph if graph else nx.MultiDiGraph()

    def serialize(self):
        data = json_graph.adjacency_data(self._graph)

        data["adjacency"] = [
            sorted(outbounds, key=lambda x: x["id"]) for outbounds in data["adjacency"]
        ]

        return data

    @classmethod
    def deserialize(cls, data):
        g = json_graph.adjacency_graph(json_util.deepcopy(data), directed=True, multigraph=True)
        return cls(graph=g)

    @staticmethod
    def get_root_nodes(graph):
        nodes = [
            {"id": n, "name": graph.node[n].get("name", n)}
            for n, d in graph.in_degree().items()
            if d == 0
        ]

        return sorted(nodes, key=lambda x: x["id"])

    @property
    def roots(self):
        return self.get_root_nodes(self._graph)

    @property
    def leaves(self):
        # Reverse the graph using a copy to identify the root nodes.
        return self.get_root_nodes(self._graph.reverse(copy=True))

    def has_tasks(self):
        return len(self._graph) > 0

    def has_task(self, task_id):
        return self._graph.has_node(task_id)

    def get_task(self, task_id):
        if not self.has_task(task_id):
            raise exc.InvalidTask(task_id)

        task = {"id": task_id}
        task.update(json_util.deepcopy(self._graph.node[task_id]))

        return task

    def get_task_attributes(self, attribute):
        return dict_util.merge_dicts(
            {n: None for n in self._graph.nodes()},
            nx.get_node_attributes(self._graph, attribute),
            overwrite=True,
        )

    def add_task(self, task_id, **kwargs):
        if not self.has_task(task_id):
            self._graph.add_node(task_id, **kwargs)
        else:
            self.update_task(task_id, **kwargs)

    def update_task(self, task_id, **kwargs):
        if not self.has_task(task_id):
            raise exc.InvalidTask(task_id)

        for key, value in six.iteritems(kwargs):
            self._graph.node[task_id][key] = value

    def has_transition(self, source, destination, **kwargs):
        edges = filter(
            lambda e: e[0] == source and e[1] == destination,
            self._graph.edges(data=True, keys=True),  # pylint: disable=unexpected-keyword-arg
        )

        for attr, value in six.iteritems(kwargs):
            edges = filter(lambda e: e[3].get(attr, None) == value, list(edges))

        return list(edges)

    def get_transition(self, source, destination, key=None, **kwargs):
        if key is not None:
            edges = filter(
                lambda e: e[0] == source and e[1] == destination and e[2] == key,
                self._graph.edges(data=True, keys=True),  # pylint: disable=unexpected-keyword-arg
            )
        else:
            edges = filter(
                lambda e: e[0] == source and e[1] == destination,
                self._graph.edges(data=True, keys=True),  # pylint: disable=unexpected-keyword-arg
            )

            for attr, value in six.iteritems(kwargs):
                edges = filter(lambda e: e[3].get(attr, None) == value, list(edges))

        edges = list(edges)

        if len(edges) <= 0:
            raise exc.InvalidTaskTransition(source, destination)

        if len(edges) > 1:
            raise exc.AmbiguousTaskTransition(source, destination)

        return edges[0]

    def get_transition_attributes(self, attribute):
        return nx.get_edge_attributes(self._graph, attribute)

    def add_transition(self, source, destination, **kwargs):
        if not self.has_task(source):
            self.add_task(source)

        if not self.has_task(destination):
            self.add_task(destination)

        # Add attributes only if value is not None.
        attrs = {}

        for attr, value in six.iteritems(kwargs):
            if attr is not None:
                attrs[attr] = value

        self._graph.add_edge(source, destination, **attrs)

    def update_transition(self, source, destination, key, **kwargs):
        seq = self.get_transition(source, destination, key=key)

        for attr, value in six.iteritems(kwargs):
            self._graph[source][destination][seq[2]][attr] = value

    def get_next_transitions(self, task_id):
        return sorted(
            [e for e in self._graph.out_edges([task_id], data=True, keys=True)], key=lambda x: x[1]
        )

    def get_prev_transitions(self, task_id):
        return sorted(
            [e for e in self._graph.in_edges([task_id], data=True, keys=True)], key=lambda x: x[1]
        )

    def get_barriers(self):
        return {
            x[0]: x[1] for x in filter(lambda x: x[1].get("barrier"), self._graph.nodes(data=True))
        }

    def set_barrier(self, task_id, value="*"):
        self.update_task(task_id, barrier=value)

    def get_barrier(self, task_id):
        return self.get_task(task_id).get("barrier")

    def has_barrier(self, task_id):
        b = self.get_barrier(task_id)

        return b is not None and b != ""

    def get_cycles(self):
        return [
            {"tasks": sorted(c), "route": nx.find_cycle(self._graph, c)}
            for c in nx.simple_cycles(self._graph)
        ]

    def in_cycle(self, task_id):
        return [c for c in nx.simple_cycles(self._graph) if task_id in c]

    def is_cycle_closed(self, cycle):
        # A cycle is closed, for a lack of better term, if there is no task
        # transition to any task that is not a member of the cycle.
        for task_id in cycle["tasks"]:
            for transition in self.get_next_transitions(task_id):
                if transition[1] not in cycle["tasks"]:
                    return False

        return True

    def get_task_retry_spec(self, task_id):
        return self.get_task(task_id).get("retry")

    def task_has_retry(self, task_id):
        r = self.get_task_retry_spec(task_id)

        return r is not None and isinstance(r, dict) and "count" in r



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta import exceptions as exc
from orquesta import requests
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorRerunTest(test_base.WorkflowConductorTest):
    def test_workflow_not_in_rerunable_status(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert rerun cannot happen because workflow is still running.
        self.assertRaises(
            exc.WorkflowIsActiveAndNotRerunableError, conductor.request_workflow_rerun
        )

    def test_task_not_in_rerunable_status(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3
          task3:
            action: core.noop
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Fail task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Assert workflow failed.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

        # Assert rerun cannot happen because task3 has not happened yet.
        self.assertRaises(
            exc.InvalidTaskRerunRequest,
            conductor.request_workflow_rerun,
            task_requests=[requests.TaskRerunRequest("task3", 0)],
        )

        # Assert rerun cannot happen because task4 does not exists.
        self.assertRaises(
            exc.InvalidTaskRerunRequest,
            conductor.request_workflow_rerun,
            task_requests=[requests.TaskRerunRequest("task4", 0)],
        )

    def test_rerun_from_failed_task(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% succeeded() %>
                publish: foobar="foobar"
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish: foobar="fubar"
                do: task3
          task3:
            action: core.noop

        output:
          - foobar: <% ctx().foobar %>
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Fail task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "foobar"})
        actual_task2_errors = [e for e in conductor.errors if e.get("task_id", None) == "task2"]
        self.assertGreater(len(actual_task2_errors), 0)

        # Request workflow rerun.
        conductor.request_workflow_rerun()

        # Assert workflow status is resuming and state is reset.
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)
        self.assertIsNone(conductor.get_workflow_output())
        actual_task2_errors = [e for e in conductor.errors if e.get("task_id", None) == "task2"]
        self.assertEqual(len(actual_task2_errors), 0)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task2")

        # Assert sequence of tasks is correct.
        staged_task = conductor.workflow_state.get_staged_task("task2", 0)
        self.assertTrue(staged_task["ready"])
        self.assertDictEqual(staged_task["ctxs"], {"in": [0, 1]})
        self.assertDictEqual(staged_task["prev"], {"task1__t0": 0})

        task_state = conductor.workflow_state.get_task("task2", 0)
        self.assertDictEqual(task_state["ctxs"], {"in": [0, 1]})
        self.assertDictEqual(task_state["prev"], {"task1__t0": 0})
        self.assertDictEqual(task_state["next"], {})

        task_states = [
            t
            for t in list(enumerate(conductor.workflow_state.sequence))
            if t[1]["id"] == "task2" and t[1]["route"] == 0
        ]

        self.assertEqual(len(task_states), 2)
        self.assertListEqual([t[0] for t in task_states], [1, 2])

        # Fail task2 again.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "foobar"})
        actual_task2_errors = [e for e in conductor.errors if e.get("task_id", None) == "task2"]
        self.assertGreater(len(actual_task2_errors), 0)

        # Assert there is a rerun record.
        expected_rerun_records = [[1]]
        self.assertListEqual(conductor.workflow_state.reruns, expected_rerun_records)

        # Request workflow rerun from task.
        task_rerun_req = requests.TaskRerunRequest("task2", 0)
        conductor.request_workflow_rerun(task_requests=[task_rerun_req])

        # Assert workflow status is resuming and state is reset.
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)
        self.assertIsNone(conductor.get_workflow_output())
        actual_task2_errors = [e for e in conductor.errors if e.get("task_id", None) == "task2"]
        self.assertEqual(len(actual_task2_errors), 0)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task2")

        # Assert sequence of tasks is correct.
        staged_task = conductor.workflow_state.get_staged_task("task2", 0)
        self.assertTrue(staged_task["ready"])
        self.assertDictEqual(staged_task["ctxs"], {"in": [0, 1]})
        self.assertDictEqual(staged_task["prev"], {"task1__t0": 0})

        task_state = conductor.workflow_state.get_task("task2", 0)
        self.assertDictEqual(task_state["ctxs"], {"in": [0, 1]})
        self.assertDictEqual(task_state["prev"], {"task1__t0": 0})
        self.assertDictEqual(task_state["next"], {})

        task_states = [
            t
            for t in list(enumerate(conductor.workflow_state.sequence))
            if t[1]["id"] == "task2" and t[1]["route"] == 0
        ]

        self.assertEqual(len(task_states), 3)
        self.assertListEqual([t[0] for t in task_states], [1, 2, 3])

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert workflow is still running.
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Succeed task3.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task3")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert workflow is completed.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "fubar"})

        # Assert there are two separate rerun records.
        expected_rerun_records = [[1], [2]]
        self.assertListEqual(conductor.workflow_state.reruns, expected_rerun_records)

    def test_rerun_from_succeeded_task(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% succeeded() %>
                publish: foobar="foobar"
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish: foobar="fubar"
                do: task3
          task3:
            action: core.noop

        output:
          - foobar: <% ctx().foobar %>
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Fail task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "foobar"})
        self.assertIn("task2", [e["task_id"] for e in conductor.errors])

        # Request workflow rerun from the succeeded task1 previous to the failed task2.
        task_rerun_req = requests.TaskRerunRequest("task1", 0)
        conductor.request_workflow_rerun(task_requests=[task_rerun_req])

        # Assert workflow status is resuming and state is reset.
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)
        self.assertIsNone(conductor.get_workflow_output())
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Assert sequence of tasks is correct.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task("task1", 0))
        self.assertIsNone(conductor.workflow_state.get_staged_task("task2", 0))

        task_states = [
            t
            for t in list(enumerate(conductor.workflow_state.sequence))
            if t[1]["id"] == "task1" and t[1]["route"] == 0
        ]

        self.assertEqual(len(task_states), 2)
        self.assertListEqual([t[0] for t in task_states], [0, 2])

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert workflow is still running.
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Succeed task3.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task3")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert workflow is completed.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "fubar"})

    def test_rerun_succeeded_workflow(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% succeeded() %>
                publish: foobar="foobar"
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish: foobar="fubar"

        output:
          - foobar: <% ctx().foobar %>
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert workflow is completed.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "fubar"})

        # Request workflow rerun from the succeeded task1.
        task_rerun_req = requests.TaskRerunRequest("task1", 0)
        conductor.request_workflow_rerun(task_requests=[task_rerun_req])

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Fail task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "foobar"})
        task2 = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(task2["status"], statuses.FAILED)
        self.assertIn("task2", [e["task_id"] for e in conductor.errors])

        # Request workflow rerun.
        conductor.request_workflow_rerun()

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert workflow is completed.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "fubar"})

    def test_rerun_default_with_remediated_task(self):
        wf_def = """
        version: 1.0
        vars:
          - foobar: null
          - fubar: null
        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% failed() %>
                publish: foobar="fubar"
                do: task3
              - when: <% succeeded() %>
                publish: foobar="foobar"
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish: fubar="fubar"
          task3:
            action: core.noop
        output:
          - foobar: <% ctx().foobar %>
          - fubar: <% ctx().fubar %>
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Fail task1 to trigger the remediation task.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Succeed the remediation task3.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[1]["id"], "task3")
        self.forward_task_statuses(conductor, next_tasks[1]["id"], fast_forward_success)

        # Fail task2 to stop the workflow.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        task1 = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(task1["status"], statuses.FAILED)
        self.assertIn("task1", [e["task_id"] for e in conductor.errors])
        task2 = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(task2["status"], statuses.FAILED)
        self.assertIn("task2", [e["task_id"] for e in conductor.errors])

        # Request workflow rerun. By default, since task1 is remediated, only task2 will be rerun.
        conductor.request_workflow_rerun()

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Render workflow output and assert workflow status, error, and output.
        self.assertEqual(len(conductor.get_next_tasks()), 0)
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        task1 = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(task1["status"], statuses.FAILED)
        self.assertIn("task1", [e["task_id"] for e in conductor.errors])
        task2 = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(task2["status"], statuses.SUCCEEDED)
        self.assertNotIn("task2", [e["task_id"] for e in conductor.errors])
        expected_workflow_output = {"foobar": "fubar", "fubar": "fubar"}
        self.assertDictEqual(conductor.get_workflow_output(), expected_workflow_output)

    def test_rerun_from_remediated_task(self):
        wf_def = """
        version: 1.0
        vars:
          - foobar: null
          - fubar: null
        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% failed() %>
                publish: foobar="fubar"
                do: task3
              - when: <% succeeded() %>
                publish: foobar="foobar"
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish: fubar="fubar"
          task3:
            action: core.noop
        output:
          - foobar: <% ctx().foobar %>
          - fubar: <% ctx().fubar %>
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Fail task1 to trigger the remediation task.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Succeed the remediation task3.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[1]["id"], "task3")
        self.forward_task_statuses(conductor, next_tasks[1]["id"], fast_forward_success)

        # Fail task2 to stop the workflow.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        task1 = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(task1["status"], statuses.FAILED)
        self.assertIn("task1", [e["task_id"] for e in conductor.errors])
        task2 = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(task2["status"], statuses.FAILED)
        self.assertIn("task2", [e["task_id"] for e in conductor.errors])

        # Request workflow rerun and explicitly ask the remediated task1 to rerun.
        tk1_rerun_req = requests.TaskRerunRequest("task1", 0)
        tk2_rerun_req = requests.TaskRerunRequest("task2", 0)
        conductor.request_workflow_rerun(task_requests=[tk1_rerun_req, tk2_rerun_req])

        # Succeed task1 and task2.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)
        self.assertEqual(next_tasks[1]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[1]["id"], fast_forward_success)

        # Render workflow output and assert workflow status, error, and output.
        self.assertEqual(len(conductor.get_next_tasks()), 0)
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        task1 = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(task1["status"], statuses.SUCCEEDED)
        self.assertNotIn("task1", [e["task_id"] for e in conductor.errors])
        task2 = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(task2["status"], statuses.SUCCEEDED)
        self.assertNotIn("task2", [e["task_id"] for e in conductor.errors])
        expected_workflow_output = {"foobar": "foobar", "fubar": "fubar"}
        self.assertDictEqual(conductor.get_workflow_output(), expected_workflow_output)

    def test_rerun_failed_task_with_items_and_reset(self):
        wf_def = """
        version: 1.0

        vars:
          - items: []
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - when: <% succeeded() %>
                publish:
                  - items: <% result() %>
                do: task2
          task2:
            action: core.noop

        output:
          - items: <% ctx(items) %>
        """

        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Fail some item(s) to fail the workflow.
        task_items = ["fee", "fi", "fo", "fum"]
        task_items_status = [statuses.SUCCEEDED] * 3 + [statuses.FAILED]

        for idx, task_item in enumerate(task_items):
            fast_forward_statuses = [statuses.RUNNING, task_items_status[idx]]
            self.forward_task_item_statuses(
                conductor,
                next_tasks[0]["id"],
                idx,
                fast_forward_statuses,
                result=task_item,
                accumulated_result=task_items[0 : idx + 1],
            )

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        task1 = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(task1["status"], statuses.FAILED)
        self.assertIn("task1", [e["task_id"] for e in conductor.errors if e.get("task_id")])
        self.assertDictEqual(conductor.get_workflow_output(), {"items": []})

        # Request workflow rerun.
        task_rerun_req = requests.TaskRerunRequest("task1", 0, reset_items=True)
        conductor.request_workflow_rerun(task_requests=[task_rerun_req])

        # Assert items are preserved in get next tasks.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(len(next_tasks[0]["actions"]), 4)

        # Succeed task1.
        task_items = ["fee", "fi", "fo", "fum"]
        task_items_status = [statuses.SUCCEEDED] * 4

        for idx, task_item in enumerate(task_items):
            fast_forward_statuses = [statuses.RUNNING, task_items_status[idx]]
            self.forward_task_item_statuses(
                conductor,
                next_tasks[0]["id"],
                idx,
                fast_forward_statuses,
                result=task_item,
                accumulated_result=task_items[0 : idx + 1],
            )

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Render workflow output and assert workflow status, error, and output.
        self.assertEqual(len(conductor.get_next_tasks()), 0)
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        task1 = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(task1["status"], statuses.SUCCEEDED)
        self.assertNotIn("task1", [e["task_id"] for e in conductor.errors])
        task2 = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(task2["status"], statuses.SUCCEEDED)
        self.assertNotIn("task2", [e["task_id"] for e in conductor.errors])
        expected_workflow_output = {"items": ["fee", "fi", "fo", "fum"]}
        self.assertDictEqual(conductor.get_workflow_output(), expected_workflow_output)

    def test_rerun_failed_task_with_items_and_no_reset(self):
        wf_def = """
        version: 1.0

        vars:
          - items: []
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - when: <% succeeded() %>
                publish:
                  - items: <% result() %>
                do: task2
          task2:
            action: core.noop

        output:
          - items: <% ctx(items) %>
        """

        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Fail some item(s) to fail the workflow.
        task_items = ["fee", "fi", "fo", "fum"]
        task_items_status = [statuses.SUCCEEDED] * 3 + [statuses.FAILED]

        for idx, task_item in enumerate(task_items):
            fast_forward_statuses = [statuses.RUNNING, task_items_status[idx]]
            self.forward_task_item_statuses(
                conductor,
                next_tasks[0]["id"],
                idx,
                fast_forward_statuses,
                result=task_item,
                accumulated_result=task_items[0 : idx + 1],
            )

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        task1 = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(task1["status"], statuses.FAILED)
        self.assertIn("task1", [e["task_id"] for e in conductor.errors if e.get("task_id")])
        self.assertDictEqual(conductor.get_workflow_output(), {"items": []})

        # Request workflow rerun.
        conductor.request_workflow_rerun()

        # Assert only 1 item is rendered in get next tasks.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(len(next_tasks[0]["actions"]), 1)

        # Succeed last item for task1.
        task_items = ["fee", "fi", "fo", "fum"]
        item_id = next_tasks[0]["actions"][0]["item_id"]

        self.forward_task_item_statuses(
            conductor,
            next_tasks[0]["id"],
            item_id,
            fast_forward_success,
            result=task_items[item_id],
            accumulated_result=task_items,
        )

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Render workflow output and assert workflow status, error, and output.
        self.assertEqual(len(conductor.get_next_tasks()), 0)
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        task1 = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(task1["status"], statuses.SUCCEEDED)
        self.assertNotIn("task1", [e["task_id"] for e in conductor.errors])
        task2 = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(task2["status"], statuses.SUCCEEDED)
        self.assertNotIn("task2", [e["task_id"] for e in conductor.errors])
        expected_workflow_output = {"items": ["fee", "fi", "fo", "fum"]}
        self.assertDictEqual(conductor.get_workflow_output(), expected_workflow_output)

    def test_rerun_from_failed_task_with_implicit_continue(self):
        wf_def = """
        version: 1.0

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - publish: foobar="foobar"

        output:
          - foobar: <% ctx().foobar %>
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Fail task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

        # Request workflow rerun.
        conductor.request_workflow_rerun()

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Render workflow output and assert workflow status, error, and output.
        self.assertEqual(len(conductor.get_next_tasks()), 0)
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        expected_workflow_output = {"foobar": "foobar"}
        self.assertDictEqual(conductor.get_workflow_output(), expected_workflow_output)

    def test_rerun_from_multiple_sequential_tasks(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% succeeded() %>
                publish: foobar="foobar"
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish: foobar="fubar"

        output:
          - foobar: <% ctx().foobar %>
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Fail task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "foobar"})
        self.assertIn("task2", [e["task_id"] for e in conductor.errors])

        # Request workflow rerun from both the succeeded task1 and the failed task2.
        tk1_rerun_req = requests.TaskRerunRequest("task1", 0)
        tk2_rerun_req = requests.TaskRerunRequest("task2", 0)
        conductor.request_workflow_rerun(task_requests=[tk1_rerun_req, tk2_rerun_req])

        # Assert workflow status is resuming and state is reset. We expect only task1
        # is setup to rerun since task2 is dependent on task1.
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)
        self.assertIsNone(conductor.get_workflow_output())
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Assert sequence of tasks is correct.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task("task1", 0))
        self.assertIsNone(conductor.workflow_state.get_staged_task("task2", 0))

        task_states = [
            t
            for t in list(enumerate(conductor.workflow_state.sequence))
            if t[1]["id"] == "task1" and t[1]["route"] == 0
        ]

        self.assertEqual(len(task_states), 2)
        self.assertListEqual([t[0] for t in task_states], [0, 2])

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert workflow is completed.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "fubar"})

    def test_rerun_with_duplicate_task_requests(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% succeeded() %>
                publish: foobar="foobar"
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish: foobar="fubar"

        output:
          - foobar: <% ctx().foobar %>
        """

        fast_forward_failure = [statuses.RUNNING, statuses.FAILED]
        fast_forward_success = [statuses.RUNNING, statuses.SUCCEEDED]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Fail task2.
        next_tasks = conductor.get_next_tasks()
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_failure)

        # Render workflow output and assert workflow status, error, and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "foobar"})
        self.assertIn("task2", [e["task_id"] for e in conductor.errors])

        # Request workflow rerun from the succeeded task1. Request task1 twice.
        tk1_rerun_req = requests.TaskRerunRequest("task1", 0)
        tk2_rerun_req = requests.TaskRerunRequest("task1", 0)
        conductor.request_workflow_rerun(task_requests=[tk1_rerun_req, tk2_rerun_req])

        # Assert workflow status is resuming and state is reset. We expect only
        # a single instance of task1 is setup.
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)
        self.assertIsNone(conductor.get_workflow_output())
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Assert sequence of tasks is correct.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task("task1", 0))
        self.assertIsNone(conductor.workflow_state.get_staged_task("task2", 0))

        task_states = [
            t
            for t in list(enumerate(conductor.workflow_state.sequence))
            if t[1]["id"] == "task1" and t[1]["route"] == 0
        ]

        self.assertEqual(len(task_states), 2)
        self.assertListEqual([t[0] for t in task_states], [0, 2])

        # Succeed task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Succeed task2.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.forward_task_statuses(conductor, next_tasks[0]["id"], fast_forward_success)

        # Assert workflow is completed.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_output(), {"foobar": "fubar"})



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta import exceptions as exc
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorContextTest(test_base.WorkflowConductorTest):
    def test_bad_app_ctx_references(self):
        wf_def = """
        version: 1.0

        input:
          - a: <% ctx().x %>

        vars:
          - b: <% ctx().y %>

        output:
          - x: <% ctx().a %>
          - y: <% ctx().b %>
          - z: <% ctx().z %>

        tasks:
          task1:
            action: core.noop
        """

        # The YaqlEvaluationException on variables "x" and "y" are a result of using unassigned
        # variables in the input and vars section. The YaqlEvaluationException on variables
        # "a", "b", and "z" are a result of rendering workflow output with unassigned variables
        # on workflow execution completion.
        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to resolve key 'x' in "
                    "expression '<% ctx().x %>' from context."
                ),
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to resolve key 'y' in "
                    "expression '<% ctx().y %>' from context."
                ),
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to resolve key 'a' in "
                    "expression '<% ctx().a %>' from context."
                ),
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to resolve key 'b' in "
                    "expression '<% ctx().b %>' from context."
                ),
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to resolve key 'z' in "
                    "expression '<% ctx().z %>' from context."
                ),
            },
        ]

        spec = native_specs.WorkflowSpec(wf_def)

        # Run the workflow.
        conductor = conducting.WorkflowConductor(spec)

        self.assertRaises(
            exc.InvalidWorkflowStatusTransition, conductor.request_workflow_status, statuses.RUNNING
        )

        # Render workflow output and check workflow status and result.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertIsNone(conductor.get_workflow_output())

    def test_app_ctx_references(self):
        app_ctx = {"x": "foobar", "y": "fubar", "z": "phobar"}

        wf_def = """
        version: 1.0

        input:
          - a: <% ctx().x %>

        vars:
          - b: <% ctx().y %>

        output:
          - x: <% ctx().a %>
          - y: <% ctx().b %>
          - z: <% ctx().z %>

        tasks:
          task1:
            action: core.noop
        """

        expected_output = app_ctx
        expected_errors = []

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(app_ctx=app_ctx), {})

        # Run the workflow.
        conductor = conducting.WorkflowConductor(spec, context=app_ctx)
        conductor.request_workflow_status(statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.assertListEqual(conductor.errors, expected_errors)

        # Complete tasks
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_ctx_yaql_queries(self):
        wf_def = """
        version: 1.0

        input:
          - vms
          - vm1: <% ctx(vms).get(vm1) %>
          - vm2: <% ctx().vms.get(vm2) %>
          - vm3: <% ctx(vms)[vm3] %>
          - vm4: <% ctx().vms[vm4] %>

        vars:
          - vm1: <% ctx(vms).get(vm1) %>
          - vm2: <% ctx().vms.get(vm2) %>
          - vm3: <% ctx(vms)[vm3] %>
          - vm4: <% ctx().vms[vm4] %>

        tasks:
          task1:
            action: mock.create
            input:
              vm1: <% ctx(vms).get(vm1) %>
              vm2: <% ctx().vms.get(vm2) %>
              vm3: <% ctx(vms)[vm3] %>
              vm4: <% ctx().vms[vm4] %>
            next:
              - publish:
                  - vm1: <% ctx(vms).get(vm1) %>
                  - vm2: <% ctx().vms.get(vm2) %>
                  - vm3: <% ctx(vms)[vm3] %>
                  - vm4: <% ctx().vms[vm4] %>

        output:
          - vm1: <% ctx(vms).get(vm1) %>
          - vm2: <% ctx().vms.get(vm2) %>
          - vm3: <% ctx(vms)[vm3] %>
          - vm4: <% ctx().vms[vm4] %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

    def test_ctx_ref_private_var(self):
        app_ctx = {"__xyz": "phobar"}

        wf_def = """
        version: 1.0

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            next:
              - publish:
                  - task1: <% ctx("__xyz") %>

        output:
          - data: <% ctx("__xyz") %>
        """

        expected_inspection_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": '<% ctx("__xyz") %>',
                    "spec_path": "output[0]",
                    "schema_path": "properties.output",
                    "message": (
                        'Variable "__xyz" that is prefixed with double underscores '
                        "is considered a private variable and cannot be referenced."
                    ),
                },
                {
                    "type": "yaql",
                    "expression": '<% ctx("__xyz") %>',
                    "spec_path": "tasks.task1.next[0].publish[0]",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.next.items.properties.publish"
                    ),
                    "message": (
                        'Variable "__xyz" that is prefixed with double underscores '
                        "is considered a private variable and cannot be referenced."
                    ),
                },
            ]
        }

        expected_conducting_errors = [
            {
                "type": "error",
                "route": 0,
                "task_id": "task1",
                "task_transition_id": "continue__t0",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx(\"__xyz\") %>'. VariableInaccessibleError: The "
                    'variable "__xyz" is for internal use and inaccessible.'
                ),
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx(\"__xyz\") %>'. VariableInaccessibleError: The "
                    'variable "__xyz" is for internal use and inaccessible.'
                ),
            },
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(app_ctx=app_ctx), expected_inspection_errors)

        # Run the workflow.
        conductor = conducting.WorkflowConductor(spec, context=app_ctx)
        conductor.request_workflow_status(statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Complete tasks
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_conducting_errors)
        self.assertIsNone(conductor.get_workflow_output())

    def test_ctx_get_all(self):
        app_ctx = {"__xyz": "phobar"}

        wf_def = """
        version: 1.0

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            next:
              - publish:
                  - task1: <% ctx() %>

        output:
          - data: <% ctx() %>
        """

        # Ensure the private variables prefixed with double underscores are not included.
        expected_output = {"data": {"foobar": "fubar", "task1": {"foobar": "fubar"}}}
        expected_errors = []

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(app_ctx=app_ctx), {})

        # Run the workflow.
        conductor = conducting.WorkflowConductor(spec, context=app_ctx)
        conductor.request_workflow_status(statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Complete tasks
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import statuses
from orquesta.tests.unit.conducting.native import base


class WorkflowErrorHandlingConductorTest(base.OrchestraWorkflowConductorTest):
    def test_error_log_fail(self):
        wf_name = "error-log-fail"

        expected_task_seq = ["task1", "log", "fail"]

        mock_statuses = [statuses.FAILED, statuses.SUCCEEDED]  # task1  # log

        mock_results = [
            "All your base are belong to us!",  # task1
            "All your base are belong to us!",  # log
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            mock_results=mock_results,
            expected_workflow_status=statuses.FAILED,
        )

    def test_error_concurrent_log_fail(self):
        wf_name = "error-log-fail-concurrent"

        expected_task_seq = ["task1", "fail", "log"]

        mock_statuses = [statuses.FAILED]  # task1

        mock_results = ["All your base are belong to us!"]  # task1

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            mock_results=mock_results,
            expected_workflow_status=statuses.FAILED,
        )

    def test_error_continue(self):
        wf_name = "error-handling-continue"

        # Run thru the success path.
        expected_routes = [
            [],  # default from start
            ["task2__t0"],  # task1 -> task2 -> continue
        ]

        expected_task_seq = ["task1", "task2", ("continue", 1)]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # continue
        ]

        expected_output = {"message": "hooray!!!"}

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=expected_routes,
            expected_output=expected_output,
        )

        # Run thru the failure path.
        expected_routes = [
            [],  # default from start
            ["task1__t0"],  # task1 -> continue
        ]

        expected_task_seq = ["task1", ("continue", 1)]

        mock_statuses = [statuses.FAILED, statuses.SUCCEEDED]  # task1  # continue

        expected_output = {"message": "$%#&@#$!!!"}

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=expected_routes,
            expected_workflow_status=statuses.FAILED,
            expected_output=expected_output,
        )

    def test_error_noop(self):
        wf_name = "error-handling-noop"

        # Run thru the success path.
        expected_task_seq = ["task1", "task2", "continue"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # continue
        ]

        expected_output = {"message": "hooray!!!"}

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name, expected_task_seq, mock_statuses=mock_statuses, expected_output=expected_output
        )

        # Run thru the failure path.
        expected_task_seq = ["task1", "noop"]

        mock_statuses = [statuses.FAILED, statuses.SUCCEEDED]  # task1  # noop

        expected_output = {"message": "$%#&@#$!!!"}

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name, expected_task_seq, mock_statuses=mock_statuses, expected_output=expected_output
        )

    def test_error_fail(self):
        wf_name = "error-handling-fail"

        # Run thru the success path.
        expected_task_seq = ["task1", "task2", "continue"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # continue
        ]

        expected_output = {"message": "hooray!!!"}

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name, expected_task_seq, mock_statuses=mock_statuses, expected_output=expected_output
        )

        # Run thru the failure path.
        expected_task_seq = ["task1", "task3", "fail"]

        mock_statuses = [
            statuses.FAILED,  # task1
            statuses.SUCCEEDED,  # task3
            statuses.FAILED,  # fail
        ]

        expected_output = {"message": "$%#&@#$!!!"}

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.FAILED,
            expected_output=expected_output,
        )



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.expressions import jinja as jinja_expr
from orquesta.tests.unit import base as test_base
from orquesta.utils import plugin as plugin_util


class JinjaEvaluationTest(test_base.ExpressionEvaluatorTest):
    @classmethod
    def setUpClass(cls):
        cls.language = "jinja"
        super(JinjaEvaluationTest, cls).setUpClass()

    def test_get_evaluator(self):
        e = plugin_util.get_module("orquesta.expressions.evaluators", self.language)

        self.assertEqual(e, jinja_expr.JinjaEvaluator)
        self.assertIn("ctx", e._custom_functions.keys())

    def test_basic_eval(self):
        expr = "{{ ctx().foo }}"

        data = {"foo": "bar"}

        self.assertEqual("bar", self.evaluator.evaluate(expr, data))

    def test_basic_eval_undefined(self):
        expr = "{{ ctx().foo }}"

        data = {}

        self.assertRaises(jinja_expr.JinjaEvaluationException, self.evaluator.evaluate, expr, data)

    def test_dict_eval(self):
        expr = "{{ ctx().nested.foo }}"

        data = {"nested": {"foo": "bar"}}

        self.assertEqual("bar", self.evaluator.evaluate(expr, data))

    def test_multi_eval(self):
        expr = "{{ ctx().foo }} and {{ ctx().marco }}"

        data = {"foo": "bar", "marco": "polo"}

        self.assertEqual("bar and polo", self.evaluator.evaluate(expr, data))

    def test_eval_recursive(self):
        expr = "{{ ctx().fee }}"

        data = {
            "fee": "{{ ctx().fi }}",
            "fi": "{{ ctx().fo }}",
            "fo": "{{ ctx().fum }}",
            "fum": "fee-fi-fo-fum",
        }

        self.assertEqual("fee-fi-fo-fum", self.evaluator.evaluate(expr, data))

    def test_eval_recursive_undefined(self):
        expr = "{{ ctx().fee }}"

        data = {"fee": "{{ ctx().fi }}", "fi": "{{ ctx().fo }}", "fo": "{{ ctx().fum }}"}

        self.assertRaises(jinja_expr.JinjaEvaluationException, self.evaluator.evaluate, expr, data)

    def test_multi_eval_recursive(self):
        expr = "{{ ctx().fee }} {{ ctx().im }}"

        data = {
            "fee": "{{ ctx().fi }}",
            "fi": "{{ ctx().fo }}",
            "fo": "{{ ctx().fum }}",
            "fum": "fee-fi-fo-fum!",
            "im": "{{ ctx().hungry }}",
            "hungry": "i'm hungry!",
        }

        self.assertEqual("fee-fi-fo-fum! i'm hungry!", self.evaluator.evaluate(expr, data))

    def test_type_preservation(self):
        data = {"k1": 101, "k2": 1.999, "k3": True, "k4": [1, 2], "k5": {"k": "v"}, "k6": None}

        self.assertEqual(data["k1"], self.evaluator.evaluate("{{ ctx().k1 }}", data))

        self.assertEqual(data["k2"], self.evaluator.evaluate("{{ ctx().k2 }}", data))

        self.assertTrue(self.evaluator.evaluate("{{ ctx().k3 }}", data))

        self.assertListEqual(data["k4"], self.evaluator.evaluate("{{ ctx().k4 }}", data))

        self.assertDictEqual(data["k5"], self.evaluator.evaluate("{{ ctx().k5 }}", data))

        self.assertIsNone(self.evaluator.evaluate("{{ ctx().k6 }}", data))

    def test_type_string_detection(self):
        expr = "{{ ctx().foo }} -> {{ ctx().bar }}"

        data = {"foo": 101, "bar": 201}

        self.assertEqual("101 -> 201", self.evaluator.evaluate(expr, data))
